/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __LIST_SKELETONS_COMMON_TPP__
#define __LIST_SKELETONS_COMMON_TPP__
#include "transporter.h"
namespace sketo {

  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_safe<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe();
    const int RE = as.get_right_unsafe();
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      int i = 0;
      for (; i < LE && i < ls; ++i) {
        comp.receive(i, as.local_get(i, l));
      }
      for (; i < ls - RE; ++i) {
        comp.receive(i, as.local_get_safe(i, l));
      }
      for (; i < ls; ++i) {
        comp.receive(i, as.local_get(i, l));
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_safe_quick_fixed<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe();
    const int RE = as.get_right_unsafe();
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      int i = 0;
      as.fix_buffer(l);
      for (; i < LE && i < ls; ++i) {
        comp.receive(i, as.local_get(i, l));
      }
      for (; i < ls - RE; ++i) {
        comp.receive(i, as.local_get_safe_quick_fixed(i));
      }
      for (; i < ls; ++i) {
        comp.receive(i, as.local_get(i, l));
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_default<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    if(as.did_zip_materialization()) {
      _impl::ls_adpt::dist_list_driver_safe<COMP, AS>::drive(comp, as);
    } else {
      _impl::ls_adpt::dist_list_driver_safe_quick_fixed<COMP, AS>::drive(comp, as);
    }    
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver(COMP &comp, const AS &as)
  {
    _impl::ls_adpt::driver_dispacher<COMP, AS>::driver::drive(comp, as);
  }

  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_ow_safe<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe() + 1;
    const int RE = as.get_right_unsafe();
    if(LE==1) {
      _impl::ls_adpt::dist_list_driver_safe<COMP, AS>::drive(comp, as);
      return;
    }
    typename AS::element_type buf[LE];
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      int i = 0;
      if(ls < LE) {
        for (; i < ls; ++i) {
          buf[i] = as.local_get(i, l);
        }
        for (i = 0; i < ls; ++i) {
          comp.receive(i, buf[i]);
        }
      } else {
        for (; i < LE; ++i) {
          buf[i] = as.local_get(i, l);
        }
        for (; i < ls - RE; ++i) {
          comp.receive(i-LE, buf[i%LE]);
          buf[i%LE] = as.local_get_safe(i, l);
        }
        for (; i < ls; ++i) {
          comp.receive(i-LE, buf[i%LE]);
          buf[i%LE] = as.local_get(i, l);
        }
        for (; i < ls + LE; ++i) {
          comp.receive(i-LE, buf[i%LE]);
        }
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_ow_safe_quick_fixed<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe() + 1;
    const int RE = as.get_right_unsafe();
    if(LE==1) {
      _impl::ls_adpt::dist_list_driver_safe_quick_fixed<COMP, AS>::drive(comp, as);
      return;
    }
    typename AS::element_type buf[LE];
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      int i = 0;
      as.fix_buffer(l);
      if(ls < LE) {
        for (; i < ls; ++i) {
          buf[i] = as.local_get(i, l);
        }
        for (i = 0; i < ls; ++i) {
          comp.receive(i, buf[i]);
        }
      } else {
        for (; i < LE; ++i) {
          buf[i] = as.local_get(i, l);
        }
        for (; i < ls - RE; ++i) {
          comp.receive(i-LE, buf[i%LE]);
          buf[i%LE] = as.local_get_safe_quick_fixed(i);
        }
        for (; i < ls; ++i) {
          comp.receive(i-LE, buf[i%LE]);
          buf[i%LE] = as.local_get(i, l);
        }
        for (; i < ls + LE; ++i) {
          comp.receive(i-LE, buf[i%LE]);
        }
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_ow_default<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    if(as.did_zip_materialization()) {
      _impl::ls_adpt::dist_list_driver_ow_safe<COMP, AS>::drive(comp, as);
    } else {
      _impl::ls_adpt::dist_list_driver_ow_safe_quick_fixed<COMP, AS>::drive(comp, as);
    }    
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_ow(COMP &comp, const AS &as)
  {
    _impl::ls_adpt::driver_ow_dispacher<COMP, AS>::driver::drive(comp, as);
  }


  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev_safe<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe();
    const int RE = as.get_right_unsafe();
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      int i = ls - 1;
      for (; i >= ls - RE && i > 0; i--) {
        comp.receive(i, as.local_get(i, l));
      }
      for (; i >= LE; i--) {
        comp.receive(i, as.local_get_safe(i, l));
      }
      for (; i >= 0; i--) {
        comp.receive(i, as.local_get(i, l));
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev_safe_quick_fixed<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe();
    const int RE = as.get_right_unsafe();
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      int i = ls - 1;
      as.fix_buffer(l);
      for (; i >= ls - RE && i > 0; i--) {
        comp.receive(i, as.local_get(i, l));
      }
      for (; i >= LE; i--) {
        comp.receive(i, as.local_get_safe_quick_fixed(i));
      }
      for (; i >= 0; i--) {
        comp.receive(i, as.local_get(i, l));
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev_default<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    if(as.did_zip_materialization()) {
      _impl::ls_adpt::dist_list_driver_rev_safe<COMP, AS>::drive(comp, as);
    } else {
      _impl::ls_adpt::dist_list_driver_rev_safe_quick_fixed<COMP, AS>::drive(comp, as);
    }    
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev(COMP &comp, const AS &as)
  {
    _impl::ls_adpt::driver_rev_dispacher<COMP, AS>::driver::drive(comp, as);
  }

  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev_ow_safe<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe();
    const int RE = as.get_right_unsafe() + 1;
    if(RE==1) {
      _impl::ls_adpt::dist_list_driver_rev_safe<COMP, AS>::drive(comp, as);
      return;
    }
    typename AS::element_type buf[RE];
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      int i = ls - 1;
      if(ls < RE) {
        for (; i >= 0; i--) {
          buf[i%RE] = as.local_get(i, l);
        }
        for (i = ls - 1; i >= 0; i--) {
          comp.receive(i, buf[i%RE]);
        }
      } else {
        for (; i >= ls - RE; i--) {
          buf[i%RE] = as.local_get(i, l);
        }
        for (; i >= LE; i--) {
          comp.receive(i+RE, buf[i%RE]);
          buf[i%RE] = as.local_get_safe(i, l);
        }
        for (; i >= 0; i--) {
          comp.receive(i+RE, buf[i%RE]);
          buf[i%RE] = as.local_get(i, l);
        }
        i+=RE;
        for (; i >= 0; i--) {
          comp.receive(i, buf[i%RE]);
        }
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev_ow_safe_quick_fixed<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    const int LE = as.get_left_unsafe();
    const int RE = as.get_right_unsafe() + 1;
    if(RE==1) {
      _impl::ls_adpt::dist_list_driver_rev_safe_quick_fixed<COMP, AS>::drive(comp, as);
      return;
    }
    typename AS::element_type buf[RE];
    for(int l = 0; l < as.get_num_bufs(); l++) {
      const int ls = as.get_local_size(l);
      comp.init(l, ls);
      as.fix_buffer(l);
      int i = ls - 1;
      if(ls < RE) {
        for (; i >= 0; i--) {
          buf[i%RE] = as.local_get(i, l);
        }
        for (i = ls - 1; i >= 0; i--) {
          comp.receive(i, buf[i%RE]);
        }
      } else {
        for (; i >= ls - RE; i--) {
          buf[i%RE] = as.local_get(i, l);
        }
        for (; i >= LE; i--) {
          comp.receive(i+RE, buf[i%RE]);
          buf[i%RE] = as.local_get_safe_quick_fixed(i);
        }
        for (; i >= 0; i--) {
          comp.receive(i+RE, buf[i%RE]);
          buf[i%RE] = as.local_get(i, l);
        }
        i+=RE;
        for (; i >= 0; i--) {
          comp.receive(i, buf[i%RE]);
        }
      }
      comp.fin(l);
    }
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev_ow_default<COMP, AS>::drive(COMP &comp, const AS &as)
  {
    if(as.did_zip_materialization()) {
      _impl::ls_adpt::dist_list_driver_rev_ow_safe<COMP, AS>::drive(comp, as);
    } else {
      _impl::ls_adpt::dist_list_driver_rev_ow_safe_quick_fixed<COMP, AS>::drive(comp, as);
    }    
  }
  template <typename COMP, typename AS>
  void
  _impl::ls_adpt::dist_list_driver_rev_ow(COMP &comp, const AS &as)
  {
    _impl::ls_adpt::driver_rev_ow_dispacher<COMP, AS>::driver::drive(comp, as);
  }

  template <typename F, typename OPLUS, typename A>
  struct map_reduce_comp {
    A val;
    std::vector<A> vals;
    const OPLUS &oplus;
    const A &identity;
    const F &f;
    map_reduce_comp(const F&f, const OPLUS &oplus, const A&identity) : f(f), oplus(oplus), identity(identity) { }
    inline void init(int , int ls) {
      val = identity;
    }
    inline void receive(int , const A& a) {
      val = oplus(val, f(a));
    }
    inline void fin(int ) {
      vals.push_back(val);
    }
  };
  template <typename OPLUS, typename A>
  struct reduce_comp : public map_reduce_comp<sketo::functions::identity<A>, OPLUS, A> {
    reduce_comp(const OPLUS &oplus, const A&identity) : map_reduce_comp<sketo::functions::identity<A>, OPLUS, A>(sketo::functions::identity<A>(), oplus, identity) { }
  };
  /*
   * This function performs "global reduce" on the values specified as pval.
   * @param oplus   An associative operator used in reduce. This opertor should be
   *                a function object of type "A oplus(A, A)".
   * @param identiy The identity element of oplus.
   * @param s       The number of reduced values of the sublists. 
   * @param pval    The address of the values reduced on each process.
   *                The result of the global reduce will be stored on this.
   */
  template <typename A, typename OP>
  void
  _impl::ls_adpt::global_reduce(const OP& oplus,
                                const A &identity,
                                int s,
                                A* pval)
  {
    A *received_data = new A[s];
    for (int stage = 1; stage < sketo::procs; stage <<= 1) {
      int target = sketo::rank ^ stage;
      if (target < sketo::rank) {
        _impl::transporter::send<A>(pval, s, target, TAG_REDUCE);
        break;
      }
      else if (target < sketo::procs) {
        _impl::transporter::recv<A>(received_data, s, target, TAG_REDUCE);
        for(int l = 0; l < s; l++) {
          pval[l] = oplus(pval[l], received_data[l]);
        }
      }
      else {
        // do nothing
      }
    }

    for(int l = 1; l < s; l++) {
      pval[0] = oplus(pval[0], pval[l]);
    }
    _impl::transporter::bcast<A>(pval, 1, 0);
    delete [] received_data;
  }

  /*
   * This function performs "global scan" on the values specified as pfirst
   * and ptotal. The first value (pfirst) is assumed to be a dummy value.
   * @param oplus   An associative operator used in scan. This operator should be
   *                a function object of type "A oplus(A, A)".
   * @param identiy The identity element of oplus.
   * @param s       The number of reduced values of the sublists. 
   * @param pfirst  The address of the values on which the result of scan is stored.
   * @param ptotal  The address of the input values that are results of local reduce.
   */
  template <typename A, typename OP>
  void
  _impl::ls_adpt::global_scan(const OP& oplus,
                              const A&identity,
                              int s,
                              A* pfirst,
                              A* ptotal)
  {
    for(int l = 0; l < s; l++) {
      pfirst[l] = identity;
    }

    A *received_val = new A[s];
    for (int stage = 1; stage < sketo::procs; stage <<= 1) {
      const int target = sketo::rank ^ stage;
      if (target >= sketo::procs) continue;
      _impl::transporter::sendrecv<A>(ptotal, received_val, s, target, TAG_SCAN1);
      if (target > sketo::rank) {
        for(int l = 0; l < s; l++) ptotal[l] = oplus(ptotal[l], received_val[l]);
      } else {
        for(int l = 0; l < s; l++) {
          ptotal[l] = oplus(received_val[l], ptotal[l]);
          pfirst[l] = oplus(received_val[l], pfirst[l]);
        }
      }
    }

    _impl::transporter::bcast<A>(ptotal, s, 0);

    // now, everyone has valid ptoral
    for(int l = 1; l < s; l++) {
      ptotal[l] = oplus(ptotal[l-1], ptotal[l]);
      pfirst[l] = oplus(ptotal[l-1], pfirst[l]);
    }
    delete [] received_val;
  }

  /*
   * This function performs "global scanr" on the values specified as plast
   * and ptotal. The last value (plast) is assumed to be a dummy value.
   * @param oplus   An associative operator used in scan. This operator should be
   *                a function object of type "A oplus(A, A)".
   * @param identiy The identity element of oplus.
   * @param s       The number of reduced values of the sublists. 
   * @param plast   The address of the values on which the result of scan is stored.
   * @param ptotal  The address of the input values that are results of local reduce.
   */
  template <typename A, typename OP>
  void
  _impl::ls_adpt::global_scanr(const OP& oplus,
                               const A&identity,
                               int s,
                               A* plast,
                               A* ptotal)
  {
    for(int l = 0; l < s; l++) {
      plast[l] = identity;
    }

    const int vrank = sketo::procs - sketo::rank - 1;
    
    A *received_val = new A[s];
    for (int stage = 1; stage < sketo::procs; stage <<= 1) {
      const int vtarget = vrank ^ stage;
      const int target = sketo::procs - vtarget - 1;
      if (target < 0) continue;
      _impl::transporter::sendrecv<A>(ptotal, received_val, s, target, TAG_SCAN1);
      if (vtarget > vrank) {
        for(int l = 0; l < s; l++) ptotal[l] = oplus(received_val[l], ptotal[l]);
      } else {
        for(int l = 0; l < s; l++) {
          ptotal[l] = oplus(ptotal[l], received_val[l]);
          plast[l] = oplus(plast[l], received_val[l]);
        }
      }
    }

    _impl::transporter::bcast<A>(ptotal, s, sketo::procs - 1);

    // now, everyone has valid ptoral
    for(int l = s - 2; l >= 0; l--) {
      ptotal[l] = oplus(ptotal[l], ptotal[l+1]);
      plast[l] = oplus(plast[l], ptotal[l+1]);
    }

    delete [] received_val;
  }
  
  template <typename OPLUS, typename A>
  struct postscan_comp {
    _impl::dist_list_buffer<A> *data;
    const std::vector<A> &total_sums;
    A total_sum;
    const OPLUS &oplus;
    const A &identity;
    postscan_comp(_impl::dist_list_buffer<A>*data, const OPLUS &oplus, const A &identity, const std::vector<A> &total_sums) : data(data), oplus(oplus), identity(identity), total_sums(total_sums) { }
    A*buf;
    inline void init(int l, int ls) {
      data->bufs.push_back(BUFFER<A>());
      data->bufs.back().resize(ls);
      buf = &(data->bufs.back()[0]);
      total_sum = total_sums[l];
    }
    inline void receive(int i, const A& a) {
      total_sum = oplus(total_sum, a);
      buf[i] = total_sum;
    }
    inline void fin(int) { }
  };

  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::postscan(const OPLUS &oplus,
                           const A &identity,
                           const dist_list<A> &as)
  {

    // local reduce
    reduce_comp<OPLUS, A> comp(oplus, identity);
    _impl::ls_adpt::dist_list_driver(comp, as);
    std::vector<A> total_sums = comp.vals;
    
    // global scan
    std::vector<A> first_vals(total_sums.size());
    global_scan(oplus, identity, total_sums.size(), &(*first_vals.begin()), &(*total_sums.begin()));

    dist_list<A> bs(0);
    bs.data->bufs.clear();
    // local scan
    {
      postscan_comp<OPLUS, A> comp(bs.data, oplus, identity, first_vals);
      _impl::ls_adpt::dist_list_driver(comp, as);
    }
    bs.data->sizes = as.data->sizes;
    bs.data->dirty = as.data->dirty;
    bs.data->global_size = as.data->global_size;
    return bs;
  }

  template <typename OPLUS, typename A>
  struct postscanr_comp {
    _impl::dist_list_buffer<A> *data;
    const std::vector<A> &total_sums;
    A total_sum;
    const OPLUS &oplus;
    const A &identity;
    postscanr_comp(_impl::dist_list_buffer<A>*data, const OPLUS &oplus, const A &identity, const std::vector<A> &total_sums) : data(data), oplus(oplus), identity(identity), total_sums(total_sums) { }
    A*buf;
    inline void init(int l, int ls) {
      data->bufs.push_back(BUFFER<A>());
      data->bufs.back().resize(ls);
      buf = &(data->bufs.back()[0]);
      total_sum = total_sums[l];
    }
    inline void receive(int i, const A& a) {
      total_sum = oplus(a, total_sum);
      buf[i] = total_sum;
    }
    inline void fin(int) { }
  };

  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::postscanr(const OPLUS &oplus,
                            const A &identity,
                            const dist_list<A> &as)
  {
    // local reduce
    reduce_comp<OPLUS, A> comp(oplus, identity);
    _impl::ls_adpt::dist_list_driver(comp, as);
    std::vector<A> total_sums = comp.vals;

    // global scan
    std::vector<A> last_vals(total_sums.size());
    global_scanr(oplus, identity, total_sums.size(), &(*last_vals.begin()), &(*total_sums.begin()));

    dist_list<A> bs(0);
    bs.data->bufs.clear();
    // local scan
    {
      postscan_comp<OPLUS, A> comp(bs.data, oplus, identity, last_vals);
      _impl::ls_adpt::dist_list_driver_rev(comp, as);
    }
    bs.data->sizes = as.data->sizes;
    bs.data->dirty = as.data->dirty;
    bs.data->global_size = as.data->global_size;
    return bs;
  }

  template <typename F, typename A>
  struct prescan_comp {
    _impl::dist_list_buffer<A> *data;
    const std::vector<A> &total_sums;
    A total_sum;
    const F &f;
    prescan_comp(_impl::dist_list_buffer<A>*data, const F &f, const std::vector<A> &total_sums) : data(data), f(f), total_sums(total_sums) { }
    A*buf;
    inline void init(int l, int ls) {
      data->bufs.push_back(BUFFER<A>());
      data->bufs.back().resize(ls);
      buf = &(data->bufs.back()[0]);
      total_sum = total_sums[l];
    }
    inline void receive(int i, const A& a) {
      buf[i] = total_sum;
      total_sum = f(total_sum, a);
    }
    inline void fin(int) { }
  };

  template <typename F, typename G, typename OMINUS, typename OPLUS,
            typename A, typename B, typename C>
  dist_list<B>
  _impl::ls_adpt::gscanl(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
                         const C &identity,
                         const dist_list<A> &as)
  {
    // local map and reduce
    map_reduce_comp<G, OPLUS, C> comp(g, oplus, identity);
    _impl::ls_adpt::dist_list_driver(comp, as);
    std::vector<C> total_sums = comp.vals;

    // global scan
    std::vector<C> first_vals(total_sums.size());
    global_scan(oplus, identity, total_sums.size(), &(*first_vals.begin()), &(*total_sums.begin()));

    dist_list<B> bs(0);
    std::vector<B> first_vals2(total_sums.size());
    for(int l = 0; l < total_sums.size(); l++) {
      first_vals2[l] = ominus(c, first_vals[l]);
    }
    // local scan
    {
      prescan_comp<F, B> comp(bs.data, f, first_vals2);
      _impl::ls_adpt::dist_list_driver(comp, as);
    }
    bs.data->sizes = as.data->sizes;
    bs.data->dirty = as.data->dirty;
    bs.data->global_size = as.data->global_size;
    return bs;
  }

  template <typename F, typename A>
  struct prescanr_comp {
    _impl::dist_list_buffer<A> *data;
    const std::vector<A> &total_sums;
    A total_sum;
    const F &f;
    prescanr_comp(_impl::dist_list_buffer<A>*data, const F &f, const std::vector<A> &total_sums) : data(data), f(f), total_sums(total_sums) { }
    A*buf;
    inline void init(int l, int ls) {
      data->bufs.push_back(BUFFER<A>());
      data->bufs.back().resize(ls);
      buf = &(data->bufs.back()[0]);
      total_sum = total_sums[l];
    }
    inline void receive(int i, const A& a) {
      buf[i] = total_sum;
      total_sum = f(total_sum, a);
    }
    inline void fin(int) { }
  };

  template <typename F, typename G, typename OMINUS, typename OPLUS,
            typename A, typename B, typename C>
  dist_list<B>
  _impl::ls_adpt::gscanr(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
                         const C &identity,
                         const dist_list<A>& as)
  {
    // local map and reduce
    map_reduce_comp<G, OPLUS, C> comp(g, oplus, identity);
    _impl::ls_adpt::dist_list_driver(comp, as);
    std::vector<C> total_sums = comp.vals;

    // global scan
    std::vector<C> last_vals(total_sums.size());
    global_scanr(oplus, identity, total_sums.size(), &(*last_vals.begin()), &(*total_sums.begin()));

    dist_list<B> bs(0);
    std::vector<B> last_vals2(total_sums.size());
    for(int l = 0; l < total_sums.size(); l++) {
      last_vals2[l] = ominus(c, last_vals[l]);
    }
    // local scan
    {
      prescan_comp<F, B> comp(bs.data, f, last_vals2);
      _impl::ls_adpt::dist_list_driver_rev(comp, as);
    }
    bs.data->sizes = as.data->sizes;
    bs.data->dirty = as.data->dirty;
    bs.data->global_size = as.data->global_size;
    return bs;
  }

}


#endif // __LIST_SKELETONS_COMMON_TPP__

