/**
 * A program translated from the Fortran version
 */
#include <iostream>
#include <cmath>
#include <utility>

#include "testcase.h"

void make_vector(double *ap, double *ae, double *aw, double *an, double *as, double *ab, double *at,
                 double *b)
{
  for (int m = 0; m < nmax; m++) b[m] = 2.0;

  for (int k = 0; k < kmax; k++) {
    for (int j = 0; j < jmax; j++) {
      for (int i = 0; i < imax; i++) {
        int m = i + j*imax + k*imax*jmax;
        if ((i > 0) && (i < imax-1)) b[m] += 2.0; else b[m] += 1.0;
        if ((j > 0) && (j < jmax-1)) b[m] += 2.0; else b[m] += 1.0;
        if ((k > 0) && (k < kmax-1)) b[m] += 2.0; else b[m] += 1.0;
      }
    }
  }
  *ap = 2.0;
  *ae = *aw = *an = *as = *at = *ab = 1.0;
}

void probmv(double ap, double ae, double aw, double an, double as, double at, double ab,
            double *b, double *c)
{
  const int di = 1;
  const int dj = imax;
  const int dk = imax * jmax;
  
  // k:0       j:0       i:0    
  {
    const int k = 0;
    const int j = 0;
    const int i = 0;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m]                + ae * b[m+di]
                                    + an * b[m+dj]
                                    + at * b[m+dk];
  }
  
  // k:0       j:0       i:imax 
  {
    const int k = 0;
    const int j = 0;
    const int i = imax-1;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m] + aw * b[m-di]
                                    + an * b[m+dj]
                                    + at * b[m+dk];
  }
  
  // k:0       j:0       i:other
  {
    const int k = 0;
    const int j = 0;
    for (int i = 1; i < imax-1; i++) {
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                                      + an * b[m+dj]
                                      + at * b[m+dk];
    }
  }
  
  // k:0       j:jmax    i:0    
  {
    const int k = 0;
    const int j = jmax-1;
    const int i = 0;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m]                + ae * b[m+di]
                     + as * b[m-dj]
                                    + at * b[m+dk];
  }
  
  // k:0       j:jmax    i:imax 
  {
    const int k = 0;
    const int j = jmax-1;
    const int i = imax-1;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m] + aw * b[m-di]
                     + as * b[m-dj]
                                    + at * b[m+dk];
  }
  
  // k:0       j:jmax    i:other
  {
    const int k = 0;
    const int j = jmax-1;
    for (int i = 1; i < imax-1; i++) {
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                       + as * b[m-dj]
                                      + at * b[m+dk];
    }
  }
  
  // k:0       j:other   i:0    
  {
    const int k = 0;
    for (int j = 1; j < jmax-1; j++) {
      const int i = 0;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m]                + ae * b[m+di]
                       + as * b[m-dj] + an * b[m+dj]
                                      + at * b[m+dk];
    }
  }
  
  // k:0       j:other   i:imax 
  {
    const int k = 0;
    for (int j = 1; j < jmax-1; j++) {
      const int i = imax-1;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di]
                       + as * b[m-dj] + an * b[m+dj]
                                      + at * b[m+dk];
    }
  }
  
  // k:0       j:other   i:other
  {
    const int k = 0;
    for (int j = 1; j < jmax-1; j++) {
      for (int i = 1; i < imax-1; i++) {
        int m = i + imax*j + imax*jmax*k;
        c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                         + as * b[m-dj] + an * b[m+dj]
                                        + at * b[m+dk];
      }
    }
  }
  
  // k:kmax    j:0       i:0    
  {
    const int k = kmax-1;
    const int j = 0;
    const int i = 0;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m]                + ae * b[m+di]
                                    + an * b[m+dj]
                     + ab * b[m-dk];
  }
  
  // k:kmax    j:0       i:imax 
  {
    const int k = kmax-1;
    const int j = 0;
    const int i = imax-1;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m] + aw * b[m-di]
                                    + an * b[m+dj]
                     + ab * b[m-dk];
  }
  
  // k:kmax    j:0       i:other
  {
    const int k = kmax-1;
    const int j = 0;
    for (int i = 1; i < imax-1; i++) {
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                                      + an * b[m+dj]
                       + ab * b[m-dk];
    }
  }
  
  // k:kmax    j:jmax    i:0    
  {
    const int k = kmax-1;
    const int j = jmax-1;
    const int i = 0;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m]                + ae * b[m+di]
                     + as * b[m-dj]
                     + ab * b[m-dk];
  }
  
  // k:kmax    j:jmax    i:imax 
  {
    const int k = kmax-1;
    const int j = jmax-1;
    const int i = imax-1;
    int m = i + imax*j + imax*jmax*k;
    c[m] = ap * b[m] + aw * b[m-di]
                     + as * b[m-dj]
                     + ab * b[m-dk];
  }
  
  // k:kmax    j:jmax    i:other
  {
    const int k = kmax-1;
    const int j = jmax-1;
    for (int i = 1; i < imax-1; i++) {
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                       + as * b[m-dj]
                       + ab * b[m-dk];
    }
  }
  
  // k:kmax    j:other   i:0    
  {
    const int k = kmax-1;
    for (int j = 1; j < jmax-1; j++) {
      const int i = 0;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m]                + ae * b[m+di]
                       + as * b[m-dj] + an * b[m+dj]
                       + ab * b[m-dk];
    }
  }
  
  // k:kmax    j:other   i:imax 
  {
    const int k = kmax-1;
    for (int j = 1; j < jmax-1; j++) {
      const int i = imax-1;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di]
                       + as * b[m-dj] + an * b[m+dj]
                       + ab * b[m-dk];
    }
  }
  
  // k:kmax    j:other   i:other
  {
    const int k = kmax-1;
    for (int j = 1; j < jmax-1; j++) {
      for (int i = 1; i < imax-1; i++) {
        int m = i + imax*j + imax*jmax*k;
        c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                         + as * b[m-dj] + an * b[m+dj]
                         + ab * b[m-dk];
      }
    }
  }
  
  // k:other   j:0       i:0    
  {
    for (int k = 1; k < kmax-1; k++) {
      const int j = 0;
      const int i = 0;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m]                + ae * b[m+di]
                                      + an * b[m+dj]
                       + ab * b[m-dk] + at * b[m+dk];
    }
  }
  
  // k:other   j:0       i:imax 
  {
    for (int k = 1; k < kmax-1; k++) {
      const int j = 0;
      const int i = imax - 1;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di]
                                      + an * b[m+dj]
                       + ab * b[m-dk] + at * b[m+dk];
    }
  }
  
  // k:other   j:0       i:other
  {
    for (int k = 1; k < kmax-1; k++) {
      const int j = 0;
      for (int i = 1; i < imax-1; i++) {
        int m = i + imax*j + imax*jmax*k;
        c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                                        + an * b[m+dj]
                         + ab * b[m-dk] + at * b[m+dk];
      }
    }
  }
  
  // k:other   j:jmax    i:0    
  {
    for (int k = 1; k < kmax-1; k++) {
      const int j = jmax-1;
      const int i = 0;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m]                + ae * b[m+di]
                       + as * b[m-dj]
                       + ab * b[m-dk] + at * b[m+dk];
    }
  }
  
  // k:other   j:jmax    i:imax 
  {
    for (int k = 1; k < kmax-1; k++) {
      const int j = jmax-1;
      const int i = imax-1;
      int m = i + imax*j + imax*jmax*k;
      c[m] = ap * b[m] + aw * b[m-di]
                       + as * b[m-dj]
                       + ab * b[m-dk] + at * b[m+dk];
    }
  }
  
  // k:other   j:jmax    i:other
  {
    for (int k = 1; k < kmax-1; k++) {
      const int j = jmax-1;
      for (int i = 1; i < imax-1; i++) {
        int m = i + imax*j + imax*jmax*k;
        c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                         + as * b[m-dj]
                         + ab * b[m-dk] + at * b[m+dk];
      }
    }
  }
  
  // k:other   j:other   i:0    
  {
    for (int k = 1; k < kmax-1; k++) {
      for (int j = 1; j < jmax-1; j++) {
        const int i = 0;
        int m = i + imax*j + imax*jmax*k;
        c[m] = ap * b[m]                + ae * b[m+di]
                         + as * b[m-dj] + an * b[m+dj]
                         + ab * b[m-dk] + at * b[m+dk];
      }
    }
  }
  
// k:other   j:other   i:imax   
  {
    for (int k = 1; k < kmax-1; k++) {
      for (int j = 1; j < jmax-1; j++) {
        const int i = imax-1;
        int m = i + imax*j + imax*jmax*k;
        c[m] = ap * b[m] + aw * b[m-di]
                         + as * b[m-dj] + an * b[m+dj]
                         + ab * b[m-dk] + at * b[m+dk];
      }
    }
  }
  
  // k:other   j:other   i:other
  for (int k = 1; k < kmax-1; k++) {
    for (int j = 1; j < jmax-1; j++) {
      for (int i = 1; i < imax-1; i++) {
        int m = i + imax*j + imax*jmax*k;
        c[m] = ap * b[m] + aw * b[m-di] + ae * b[m+di]
                         + as * b[m-dj] + an * b[m+dj]
                         + ab * b[m-dk] + at * b[m+dk];
      }
    }
  }

}

std::pair<int, double>
bicgstab_band(double ap, double ae, double aw ,double an, double as, double at, double ab,
              double *x, double *b)
{
//   double er0 = EPS;

  double *c = new double[nmax];
  probmv(ap, ae, aw, an, as, at, ab, x, c);

  double *r = new double[nmax];
  double c1 = 0;
  for (int m = 0; m < nmax; m++) { r[m] = b[m] - c[m]; c1 += r[m] * r[m]; }
//   if (c1 < er0) {
//     delete[] r;
//     delete[] c;
//     return std::pair<int, double>(0, c1);
//   }

  double *p  = new double[nmax]; for (int m = 0; m < nmax; m++) { p[m] = r[m]; }
  double *r0 = new double[nmax]; for (int m = 0; m < nmax; m++) { r0[m] = r[m]; }

  double *e  = new double[nmax];
  double *y  = new double[nmax];
  double *v  = new double[nmax];
  std::pair<int, double> ret;
  for (int it = 0; it < ITRMAX; it++) {
    probmv(ap, ae, aw, an, as, at, ab, p, y);
    double c2 = 0;
    for (int m = 0; m < nmax; m++) { c2 += r0[m] * y[m]; }
    double alp = c1 / c2;
    
    for (int m = 0; m < nmax; m++) { e[m] = r[m] - alp*y[m]; }
    probmv(ap, ae, aw, an, as, at, ab, e, v);
    
    double ev = 0;
    for (int m = 0; m < nmax; m++) { ev += e[m] * v[m]; }
    double vv = 0;
    for (int m = 0; m < nmax; m++) { vv += v[m] * v[m]; }
    
    double c3 = ev / vv;
    for (int m = 0; m < nmax; m++) { x[m] = x[m] + alp*p[m] + c3*e[m]; }
    for (int m = 0; m < nmax; m++) { r[m] = e[m] - c3*v[m]; }
    
    double rr = 0;
    for (int m = 0; m < nmax; m++) { rr += r[m] * r[m]; }
//     if (std::fabs(rr) < er0) { ret = std::pair<int, double>(it + 1, rr); break; }

    c1 = 0;
    for (int m = 0; m < nmax; m++) { c1 += r0[m] * r[m]; }
    double bet = c1/(c2 * c3);
    for (int m = 0; m < nmax; m++) { p[m] = r[m] + bet*(p[m]-c3*y[m]); }

    ret = std::pair<int, double>(it + 1, rr);
  }
  delete[] v;
  delete[] y;
  delete[] e;
  delete[] r0;
  delete[] p;
  delete[] r;
  delete[] c;
  return ret;
}
  
int main(int, char**)
{
  double ap, ae, aw, an, as, at, ab;
  double *b = new double[nmax];
  make_vector(&ap, &ae, &aw, &an, &as, &at, &ab, b);

  double *x = new double[nmax];
  for (int m = 0; m < nmax; ++m) { x[m] = 1.0 + 0.01; }
  
  clock_t st, et;
  st = clock();
#ifndef PROBMV_ONLY
  std::pair<int, double> result = bicgstab_band(ap, ae, aw, an, as, at, ab,
                                                x, b);
//   std::cout << "errer = " << result.second << std::endl;
#else
  ap = -4.5;
  double *y = new double[nmax_b];
  // avoid too-much optimization
  double tmpval = 0.0;
  for (int i = 0; i < ITRMAX * 2; i++) {
    probmv(ap, ae, aw, an, as, at, ab, x, y);
    tmpval += y[0];
  }
  std::cout << tmpval << std::endl;
#endif
  et = clock();
  std::cout << "time = " << (double)(et - st)/CLOCKS_PER_SEC << std::endl;
}
