#include <iostream>
#include <cmath>
#include <utility>

#include <sketo/list_skeletons.h>

namespace sl = sketo::list_skeletons;

#include "testcase.h"

struct array2D {
  double data[jmax+2][imax+2];

//   array2D& operator=(array2D x) {
//     for (int j = 0; j < jmax+2; j++) {
//       for (int i = 0; i < imax+2; i++) {
// 	data[j][i] = x.data[j][i];
//       }
//     }
//     return *this;
//   }
};

array2D zero_2D;

struct plus_2D_t : public sketo::functions::base<array2D (array2D, array2D)> {
  array2D operator()(const array2D &x, const array2D &y) const {
    array2D ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.data[j][i] = x.data[j][i] + y.data[j][i];
      }
    }
    return ret;
  }
} plus_2D;

struct times_2D_t : public sketo::functions::base<array2D (array2D, array2D)> {
  array2D operator()(const array2D &x, const array2D &y) const {
    array2D ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.data[j][i] = x.data[j][i] * y.data[j][i];
      }
    }
    return ret;
  }
} times_2D;

struct times_scalar_2D : public sketo::functions::base<array2D (array2D)> {
  double scalar;
  times_scalar_2D(double scalar) : scalar(scalar) { }
  array2D operator()(const array2D &x) const {
    array2D ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.data[j][i] = scalar * x.data[j][i];
      }
    }
    return ret;
  }
};

struct minus_2D_t : public sketo::functions::base<array2D (array2D, array2D)> {
  array2D operator()(const array2D &x, const array2D &y) const {
    array2D ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.data[j][i] = x.data[j][i] - y.data[j][i];
      }
    }
    return ret;
  }
} minus_2D;

struct sqsum_2D_t : public sketo::functions::base<double (array2D)> {
  double operator()(const array2D &x) const {
    double ret = 0.0;
    for (int j = 1; j <= jmax; j++) {
      for (int i = 1; i <= imax; i++) {
	ret += (x.data[j][i] * x.data[j][i]);
      }
    }
    return ret;
  }
} sqsum_2D;

struct sum_2D_t : public sketo::functions::base<double (array2D)> {
  double operator()(const array2D &x) const {
    double ret = 0.0;
    for (int j = 1; j <= jmax; j++) {
      for (int i = 1; i <= imax; i++) {
	ret += x.data[j][i];
      }
    }
    return ret;
  }
} sum_2D;




struct gen_b_t : public sketo::functions::base<array2D (int)> {
  array2D operator()(int k) const {
    array2D ret;
    for (int i = 0; i < imax + 2; i++) {
      for (int j = 0; j < jmax + 2; j++) {
	double v;
	if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1)) {
	  v = 0.0;
	} else {
	  v = 2.0;
	  if ((i > 1) && (i < imax)) v += 2.0; else v += 1.0;
	  if ((j > 1) && (j < jmax)) v += 2.0; else v += 1.0;
	  if ((k > 0) && (k < kmax-1)) v += 2.0; else v += 1.0;
	}
	ret.data[j][i] = v;
      }
    }
    return ret;
  }
} gen_b;

struct get_f_t : public sketo::functions::base<array2D (int)> {
  array2D operator()(int k) const {
    array2D ret;
    for (int i = 0; i < imax + 2; i++) {
      for (int j = 0; j < jmax + 2; j++) {
	if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1)) {
	  ret.data[j][i] = 0.0;
	} else {
	  ret.data[j][i] = 1.0;
	}
      }
    }
    return ret;
  }
} gen_f;
    
void make_vector(double *ap, double *ae, double *aw, double *an, double *as, double *ab, double *at,
		 sketo::dist_list<array2D> *b, sketo::dist_list<array2D> *f)
{
  *b = sl::generate(kmax, gen_b);
  *f = sl::generate(kmax, gen_f);

  *ap = 2.0;
  *ae = *aw = *an = *as = *at = *ab = 1.0;
}

struct local_2D : public sketo::functions::base<array2D (array2D)> {
  double ap, aw, ae, as, an;
  local_2D(double ap, double aw, double ae, double as, double an) :
    ap(ap), aw(aw), ae(ae), as(as), an(an) { }
  array2D operator()(const array2D &x) const {
    array2D ret;
    for (int j = 1; j <= jmax; j++) {
      for (int i = 1; i <= imax; i++) {
	ret.data[j][i] = ap * x.data[j][i] + aw * x.data[j][i-1] + ae * x.data[j][i+1]
	                                   + as * x.data[j-1][i] + an * x.data[j+1][i];
      }
    }
    return ret;
  }
};

void probmv(double ap, double ae, double aw, double an, double as, double at, double ab,
	    const sketo::dist_list<array2D> &b, sketo::dist_list<array2D> *c,
	    const sketo::dist_list<array2D> &f)
{
  *c = sl::zipwith(times_2D, f,
       sl::zipwith(plus_2D, sl::map(local_2D(ap, aw, ae, as, an), b),
		   sl::zipwith(plus_2D, sl::map(times_scalar_2D(ab), sl::shiftl(zero_2D, b)),
			       sl::map(times_scalar_2D(at), sl::shiftr(zero_2D, b)))));
}

std::pair<int, double>
bicgstab_band(double ap, double ae, double aw ,double an, double as, double at, double ab,
	      sketo::dist_list<array2D> *x,
	      sketo::dist_list<array2D> *b,
	      const sketo::dist_list<array2D> &f)
{
//   double er0 = EPS;

  sketo::dist_list<array2D> c, r, p, r0, e, y, v;

  probmv(ap, ae, aw, an, as, at, ab, *x, &c, f);

//     for (int k = 0; k < kmax; k++) {
//       array2D tmp = c.local_get(k);
//       for (int j = 1; j <= jmax; j++) {
// 	for (int i = 1; i <= imax; i++) {
// 	  std::cout << " " << tmp.data[j][i];
// 	}
//       }
//       std::cout << std::endl;
//     }
//     std::cout << std::endl;

  r = sl::zipwith(minus_2D, *b, c);
  double c1 = sl::reduce(std::plus<double>(),
			 sl::map(sqsum_2D, r));

//   if (c1 < er0) {
//     return std::pair<int, double>(0, c1);
//   }

  p = r.clone();
  r0 = r.clone();

  std::pair<int, double> ret;
  for (int it = 0; it < ITRMAX; it++) {

    probmv(ap, ae, aw, an, as, at, ab, p, &y, f);

    double c2 = sl::reduce(std::plus<double>(),
			   sl::map(sum_2D, 
				   sl::zipwith(times_2D, r0, y)));
    double alp = c1 / c2;

    e = sl::zipwith(minus_2D,
		    r,
		    sl::map(times_scalar_2D(alp), y));

    probmv(ap, ae, aw, an, as, at, ab, e, &v, f);

    double ev = sl::reduce(std::plus<double>(),
			   sl::map(sum_2D, 
				   sl::zipwith(times_2D, e, v)));
    double vv = sl::reduce(std::plus<double>(),
			   sl::map(sum_2D, 
				   sl::zipwith(times_2D, v, v)));
    
    double c3 = ev / vv;
    *x = sl::zipwith(plus_2D,
		     *x, sl::zipwith(plus_2D,
				     sl::map(times_scalar_2D(alp), p),
				     sl::map(times_scalar_2D(c3), e)));
    r = sl::zipwith(minus_2D, e, sl::map(times_scalar_2D(c3), v));
    
    double rr = sl::reduce(std::plus<double>(),
			   sl::map(sum_2D, 
				   sl::zipwith(times_2D, r, r)));
//     if (std::fabs(rr) < er0) { ret = std::pair<int, double>(it + 1, rr); break; }

    c1 = sl::reduce(std::plus<double>(),
		    sl::map(sum_2D,
			    sl::zipwith(times_2D, r0, r)));
    double bet = c1/(c2 * c3);
    p = sl::zipwith(plus_2D,
		    r, sl::zipwith(plus_2D,
				   sl::map(times_scalar_2D(bet), p),
				   sl::map(times_scalar_2D(-bet * c3), y)));
    
    ret = std::pair<int, double>(it + 1, rr);
  }
  return ret;
}

struct gen_x_t : public sketo::functions::base<array2D (int)> {
  array2D operator()(int k) const {
    array2D ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1)) {
	  ret.data[j][i] = 0.0;
	} else {
	  ret.data[j][i] = 1.0 + 0.01;
	}
      }
    }
    return ret;
  }
} gen_x;

int sketo::main(int, char**)
{
  // zero_2D
  for (int j = 0; j < jmax+2; j++) {
    for (int i = 0; i < imax+2; i++) {
      zero_2D.data[j][i] = 0.0;
    }
  }
  
  double ap, ae, aw, an, as, at, ab;
  sketo::dist_list<array2D> b, f;
  make_vector(&ap, &ae, &aw, &an, &as, &at, &ab, &b, &f);

  sketo::dist_list<array2D> x = sl::generate(kmax, gen_x);
  
  double st, et;
  MPI_Barrier(MPI_COMM_WORLD);
  st = MPI_Wtime();
#ifndef PROBMV_ONLY
  std::pair<int, double> result = bicgstab_band(ap, ae, aw, an, as, at, ab,
						&x, &b, f);
  std::cout << "error = " << result.second << std::endl;
#else
  ap = -4.5;
  sketo::dist_list<array2D> y(kmax);
  // avoid too-much optimization
  double tmpval = 0.0;
  for (int i = 0; i < ITRMAX * 2; i++) {
    probmv(ap, ae, aw, an, as, at, ab, x, &y, f);
    tmpval += sketo::_impl::dist_list_buffer<array2D>::get_buffer(y)->buf[0].data[0][0];
  }
  sketo::cout << tmpval << std::endl;
#endif
  MPI_Barrier(MPI_COMM_WORLD);
  et = MPI_Wtime();
  sketo::cout << "time = " << (et - st) << std::endl;
  return 0;
}
