/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file mcpi.cpp
 * @brief This program computes the pi using Monte-Carlo method/
 */

#include <sketo/list_skeletons.h>
#include <cstdlib>
#include <ctime>
#include <utility>

//-------------- for measuring computation time
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
double gettimeofday_sec()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}
//---------------


typedef std::pair<double, double> point;

struct ranpoint_t : public sketo::functions::base<point (int)> {
  point operator()(int /* x */) const {
    return point((std::rand() * 1.0 / RAND_MAX), (std::rand() * 1.0 / RAND_MAX));
  }
} randpoint;

struct distance_t : public sketo::functions::base<double(point)> {
  double operator()(point x) const {
    return x.first * x.first + x.second * x.second;
  }
} distance;

struct inout_t : public sketo::functions::base<double(double)> {
  double operator()(double x) const {
    return x <= 1.0 ? 1.0 : 0.0;
  }
} inout;

struct plus_t : public sketo::functions::base<double(double, double)> {
  double operator()(double x, double y) const {
    return x + y;
  }
} plus;

using namespace sketo::list_skeletons;

int sketo::main(int argc, char** argv)
{
  if (argc < 2) {
    sketo::cout << "usage: " << argv[ 0 ] << " size" << std::endl;
    return 0;
  }

  const int size = std::atoi(argv[1]);

  // this should use PID or something related to the process.
  std::srand(std::time(NULL)); // initialize random seed

  double st = gettimeofday_sec(); // start time

  dist_list<point> ps = generate(size, randpoint);
  dist_list<double> ds = map(distance, ps);
  dist_list<double> inouts = map(inout, ds);
  double pi = reduce(std::plus<double>(), inouts) / size * 4;
//  double pi = reduce(plus, inouts) / size * 4;

  // fusable!
  //double pi = reduce(std::plus<double>(), map(inout, map(distance, generate(size, randpoint)))) / size * 4;

  
  sketo::cout << pi << std::endl; // output the result
  double et = gettimeofday_sec(); // end time
  sketo::cout << (et-st) << " " << std::endl;
  return 0;
}
