/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

#include <sketo/list_skeletons.h>
#include <cstdlib>
//-------------- for measuring computation time
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string>

double gettimeofday_sec()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}
//---------------

void run_map(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), 2), x);
}

void run_zip(sketo::dist_list<double>&x)
{
  sketo::dist_list<std::pair<double, double> > y =
    sketo::list_skeletons::zip(x, x);
}

void run_reduce(sketo::dist_list<double>&x)
{
  double y =
    sketo::list_skeletons::reduce(std::plus<double>(), x);
}

void run_scan(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::scan(std::plus<double>(), 0.0, x);
}

void run_scanr(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::scanr(std::plus<double>(), 0.0, x);
}

void run_scan_scanr(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::scanr
    (std::plus<double>(), 0.0, 
     sketo::list_skeletons::scan(std::plus<double>(), 0.0, x));
}

void run_scan_scanr_map(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 2),
    sketo::list_skeletons::scanr
     (std::plus<double>(), 0.0, 
      sketo::list_skeletons::scan(std::plus<double>(), 0.0, x)));
}

void run_scan_map(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 2),
     sketo::list_skeletons::scan(std::plus<double>(), 0.0, x));
}

void run_scanr_map(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 2),
     sketo::list_skeletons::scanr(std::plus<double>(), 0.0, x));
}

void run_shiftl(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::shiftl(1.0, x);
}

void run_shiftr(sketo::dist_list<double>&x)
{
  sketo::dist_list<double> y =
    sketo::list_skeletons::shiftr(1.0, x);
}

void run_zipw_ow(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::zipwith(std::plus<double>(), x, x);
}

void run_zipw_map_ow(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::zipwith(
      std::plus<double>(),
      x,
      sketo::list_skeletons::map(
        std::bind1st(std::multiplies<double>(), 0.1),
        x));
}


void run_shiftl_ow(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::shiftl(1.0, x);
}

void run_shiftr_ow(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::shiftl(1.0, x);
}

void run_map_ow(sketo::dist_list<double>&x)
{
  x = sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), 0.99), x);
}

void run_map_ow2(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 0.99),
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 0.99),
     x));
}

void run_map_ow4(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 0.99),
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 0.99),
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 0.99),
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 0.99),
     x))));
}


void run_scan_ow(sketo::dist_list<double>&x)
{
  x = 
    sketo::list_skeletons::scan(std::plus<double>(), 0.0, x);
}

void run_scanr_ow(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::scanr(std::plus<double>(), 0.0, x);
}


void run_scan_map_ow(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 2),
     sketo::list_skeletons::scan(std::plus<double>(), 0.0, x));
}

void run_scanr_map_ow(sketo::dist_list<double>&x)
{
  x =
    sketo::list_skeletons::map
    (std::bind1st(std::multiplies<double>(), 2),
     sketo::list_skeletons::scanr(std::plus<double>(), 0.0, x));
}

void run_heat_eq(sketo::dist_list<double>&x)
{
x = sketo::list_skeletons::zipwith(
  std::plus<double>(),
  x,
  sketo::list_skeletons::map(
    std::bind1st(std::multiplies<double>(), 0.1),
    sketo::list_skeletons::zipwith(
      std::plus<double>(),
      sketo::list_skeletons::zipwith(
        std::plus<double>(),
        sketo::list_skeletons::shiftl(0.0, x),
        sketo::list_skeletons::shiftr(0.0, x)),
      sketo::list_skeletons::map(
        std::bind1st(std::multiplies<double>(), -2.0), x))));
}

struct microbench {
  microbench(void (*run)(sketo::dist_list<double>&), const char *name) : run(run), name(name) { }
  void (*run)(sketo::dist_list<double>&);
  std::string name;
};

#define MICROBENCH(x) microbench(run_##x, #x)


microbench benches [] = {
  MICROBENCH( map ),
  MICROBENCH( zip ),
  MICROBENCH( shiftl ),
  MICROBENCH( shiftr ),
  MICROBENCH( reduce ),
  MICROBENCH( scan ),
  MICROBENCH( scanr ),
  MICROBENCH( scan_map ),
  MICROBENCH( scanr_map ),
  MICROBENCH( scan_scanr ),
  MICROBENCH( scan_scanr_map ),
  
  MICROBENCH( map_ow ),
  MICROBENCH( map_ow2 ),
  MICROBENCH( map_ow4 ),
  MICROBENCH( zipw_ow ),
  MICROBENCH( shiftl_ow ),
  MICROBENCH( shiftr_ow ),
  MICROBENCH( scan_ow ),
  MICROBENCH( scanr_ow ),
  MICROBENCH( scan_map_ow ),
  MICROBENCH( scanr_map_ow ),
  
  MICROBENCH( heat_eq ),
  microbench(NULL, "")
};

void usage(const char *prg) {
  sketo::cout << "usage: " << prg << "program length times" << std::endl;
  sketo::cout << "available programs: " << std::endl;
  microbench *p;
  for(p = benches; p->run != NULL; p++) {
    sketo::cout << "   " << p->name << std::endl;
  }
}

int sketo::main(int argc, char **argv)
{
  if (argc != 4) {
    usage(argv[0]);
    exit(0);
  }
  std::string ske(argv[1]);
  const int length = std::atoi(argv[2]); // length of bar
  const int times = std::atoi(argv[3]);   // passed time

  microbench *p;
  for(p = benches; p->run != NULL; p++) {
    if(ske == p->name) break;
  }
  if(p->run == NULL) {
    usage(argv[0]);
    exit(0);
  }
  sketo::cout << "running " << p->name << " at " << sketo::procs << " procs" << "... ";
  sketo::dist_list<double> bar = sketo::list_skeletons::generate(length, sketo::functions::identity<double>());
  MPI_Barrier( MPI_COMM_WORLD );
  double st = gettimeofday_sec();
  
  for(int i = 0; i < times; i++) {
    p->run(bar);
  }
  MPI_Barrier( MPI_COMM_WORLD );
  sketo::cout << "time: " << (gettimeofday_sec() - st) <<  std::endl;
//   skeleton::time::set_time( "calculate" ); // for time keeping
#if DEBUG_OUTPUT
  list_skeletons::print( bar );
#endif

  return 0;
}

/*

mpicxx -L../../src -lsketo -I../../include -O3 microbench.cpp -o microbench

prgs=`./microbench | sed -e '1,2d'`

num_trial=10
sizes="1M-1K 10M-100"

for x in $sizes; do
  echo "#elements-#iterations = $x"
  logfileall=log-$x.txt
  rm -rf $logfileall
  for k in `seq 1 $num_trial`; do
    echo "running $k-th set"
    logfile=log-$x-$k.txt
    rm -rf $logfile
    n=`echo $x | sed -e 's|-.*||;s/M/000000/;s/K/000/'`
    t=`echo $x | sed -e 's|.*-||;s/M/000000/;s/K/000/'`
    for prg in $prgs; do
      for p in 1 2 4 8; do
        mpirun -np $p ./microbench $prg $n $t >> $logfile
      done
    done
    if test $k = 1; then
      cp $logfile $logfileall
    else
      (cat -n $logfileall; cat -n $logfile) | sort | sed -e 'N;s|^ *[0-9]*\t||;s|\n.*: \(.*\)| \1|' > __tmp
      mv __tmp $logfileall
    fi
  done
  mintime=minruntimes-of-$k-$x.txt
  cat $logfileall | sed -e 'h;s|.*:||;s| [^ ]*|echo &;|g;s|.*|(&)|;s!$! | awk "BEGIN{a=1000000}//{if(\\\$1 < a) {a = \\\$1}}END{print a}"!e;x;s|:.*|:|;G;s|\n| |' > $mintime  
  cat $mintime | sed -e 's|^running \([^ ]*\).* time: \([^ ]*\).*|\1 \2|' | sed -e 'N;N;N;s|\n[^ ]*||g' | awk '//{printf("%-20s %-8.3g %-8.3g %-8.3g %-8.3g\n", $1 ,$2 ,$3 ,$4, $5)}' > table-$x.txt
done

for x in $sizes; do 
  (cat sketo-1.10/table-$x.txt | sed -e 's/^/sketo-1.10 /' | cat -n ; cat sketo-new/table-$x.txt  | sed -e 's/^/sketo-new  /' | cat -n ) | sort | sed -e 's/^ *[0-9]*\t//g;2~2s/$/\n/' > compare-$x.txt
done


 */
