/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file mpadd.cpp
 * @brief A program for addition of multi-precision numbers
 */

#include <sketo/list_skeletons.h>
#include <cstdlib>
#include <ctime>
#include <utility>

const int base = 10;

typedef std::pair<int, int> ipair;

struct randint_t : public sketo::functions::base<int (int)> {
  int operator()(int /* x */) const {
    return (int)((std::rand() * 1.0 / RAND_MAX) * base) % base;
  }
} randint;

struct generator_propagator_t : public sketo::functions::base<ipair(int)> {
  ipair operator()(int x) const {
    return ipair(x == base-1, x >= base);
  }
} generator_propagator;

struct snd_t : public sketo::functions::base<int(ipair)> {
  int operator()(ipair x) const {
    return x.second;
  }
} snd;

// struct times_t : public sketo::functions::base<ipair(ipair, ipair)> {
//   ipair operator()(ipair x, ipair y) const {
//     return ipair(x.first && y.first, (y.first && x.second) || y.second);
//   }
// } times;
// ipair identity_element(const times_t&) {
//   return ipair(1, 0);
// }

struct timesr_t : public sketo::functions::base<ipair(ipair, ipair)> {
  ipair operator()(ipair x, ipair y) const {
    return ipair(x.first && y.first, (x.first && y.second) || x.second);
  }
} timesr;
ipair identity_element(const timesr_t&) {
  return ipair(1, 0);
}

struct addmod_t : public sketo::functions::base<int(int, int)> {
  int operator()(int x, int y) const {
    return (x + y) % base;
  }
} addmod;

template <typename A, typename B>
std::ostream& operator<<(std::ostream& os, const std::pair<A, B>& x) {
  os << "(" << x.first << "," << x.second << ")";
  return os;
}


using namespace sketo::list_skeletons;

int sketo::main(int argc, char** argv)
{
  if (argc < 2) {
    sketo::cout << "usage: " << argv[ 0 ] << " size" << std::endl;
    return 0;
  }

  const int size = std::atoi(argv[1]);

  std::srand(std::time(NULL)); // initialize random seed

  ipair e(1,0);
  ipair all;
  int carry = 0;
  
  dist_list<int> x = generate(size, randint);
  dist_list<int> y = generate(size, randint);
  dist_list<int> z = zipwith(std::plus<int>(), x, y);
  dist_list<ipair> gsps = map(generator_propagator, z);
  dist_list<int> cs = map(snd, scanr(timesr, e, gsps, &all));
  carry = snd(all);
  
//  dist_list<int> c = shiftr(0, map(snd, postscan(times, gp)));
  dist_list<int> r = zipwith(addmod, z, c);

  sketo::cout << "    x =   "; x.print();
  sketo::cout << "    y =   "; y.print();
  sketo::cout << "--------------------" << std::endl;
  sketo::cout << "x + y = " << (carry ? "1" : " " ) << " "; r.print();
 
  return 0;
}
