/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file test_matrix_skeletons.cpp
 * @brief Test code for matrix_skeletons.h and matrix_skeletons.cpp
 */

#include <mpi.h>
//#define __SKETO_NO_FUSION__

#include "../include/sketo/sketo.h"
#include "../include/sketo/matrix_skeletons.h"
//#include "../include/sketo/dist_matrix.h"

#include <cassert>
#include <utility>
#include <iostream>
#include <functional>
#include <vector>

const sketo::msize size(13, 11);

typedef std::pair<int, int> int_pair;

// struct int_pair {
//   int first;
//   int second;
//   int_pair() {}
//   int_pair(int first, int second) : first(first), second(second) { }
//   bool operator==(const int_pair &x) const { return first == x.first && second == x.second; }
//   bool operator!=(const int_pair &x) const { return !(x==*this); }
// };

// std::ostream& operator<<(std::ostream& os, const int_pair& x) {
//   os << "(" << x.first << "," << x.second << ")";
//   return os;
// }

typedef std::pair<double, double> double_pair;

typedef std::vector<int> int_vector;

typedef std::vector<int_vector> int_mat;


template <typename A, typename B>
std::ostream& operator<<(std::ostream& os, const std::pair<A, B>& x) {
  os << "(" << x.first << "," << x.second << ")";
  return os;
}

template <typename A>
std::ostream& operator<<(std::ostream& os, const std::vector<A>& x) {
  os << "[";
  for(unsigned int i = 0; i < x.size(); i++) {
    if(i > 0) os << ",";
    os << x[i];
  }
  os << "]";
  return os;
}

template <typename T>
bool same_arrays(T* as, T* bs, sketo::msize size)
{
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      if (as[i * size.j + j] != bs[i * size.j + j]) {
        std::cerr << "Wrong data: " << "LHS = " << as[i * size.j + j] << ", RHS = " << bs[i * size.j + j] << ". at index " << "(" << i << "," << j << ")" << "." << std::endl;
        return false;
      }
    }
  }
  return true;
}

class im_single_t : public sketo::functions::base <int_mat(sketo::mindex)> {
public: int_mat operator()(sketo::mindex mi) const { int x = mi.i * size.j + mi.j; int_mat im; int_vector iv; iv.push_back(x); im.push_back(iv); return im; }
} im_single;

class im_abv_t : public sketo::functions::base <int_mat(int_mat, int_mat)> {
public: int_mat operator()(int_mat x, int_mat y) const {
  int_mat im;
  for(unsigned int i = 0; i < x.size(); i++) im.push_back(x[i]);
  for(unsigned int i = 0; i < y.size(); i++) im.push_back(y[i]);  
  return im;
};
} im_abv;

class im_bsd_t : public sketo::functions::base <int_mat(int_mat, int_mat)> {
public: int_mat operator()(int_mat x, int_mat y) const {
  int_mat im;
  for(unsigned int i = 0; i < x.size(); i++) {
    int_vector iv;
    for(unsigned int j = 0; j < x[i].size(); j++) iv.push_back(x[i][j]);
    for(unsigned int j = 0; j < y[i].size(); j++) iv.push_back(y[i][j]);
    im.push_back(iv);
  }
  return im;
};
} im_bsd;



class to_int_pair_t : public sketo::functions::base <int_pair(sketo::mindex)> {
public: int_pair operator()(sketo::mindex mi) const { int x = mi.i * size.j + mi.j; return int_pair(x + 1, x * x); }
} to_int_pair;

class to_int_pair2_t : public sketo::functions::base <int_pair(sketo::mindex)> {
public: int_pair operator()(sketo::mindex mi) const { int x = mi.i * size.j + mi.j; return int_pair(x * 2 + 1, x + 3); }
} to_int_pair2;

class shift_fun_identity_t : public sketo::functions::base <int_pair(sketo::mindex, int_pair)> {
public: int_pair operator()(sketo::mindex mi, int_pair) const { return int_pair(mi.i, mi.j); }
} shift_fun_identity;


class to_int_vector_t : public sketo::functions::base <int_vector(sketo::mindex)> {
public:
  int_vector operator()(sketo::mindex mi) const {
    int_vector v;
    for(int i = 0; i <= mi.i; i++) {
      v.push_back(i + mi.j);
    }
    return v;
  }
} to_int_vector;

class doublize_t : public sketo::functions::base <double(sketo::mindex)> {
public: double operator()(sketo::mindex mi) const { double x = mi.i * size.j + mi.j; return x; }
} doublize;

class intize_t : public sketo::functions::base <int(sketo::mindex)> {
public: int operator()(sketo::mindex mi) const { int x = 1 + 0*(mi.i * size.j + mi.j); return x; }
} intize;


// function object for map
class same_last_two_bit_t : public sketo::functions::base <bool(int_pair)> {
public: bool operator()(int_pair x) const { return ((x.first & 3) == (x.second & 3)); }
} same_last_two_bit;

// function object for map_index
class same_last_two_bit_with_index_t : public sketo::functions::base <bool(sketo::mindex, int_pair)> {
public: bool operator()(sketo::mindex x, int_pair y) const { return (((x.i + y.first) & 3) == (y.second & 3)); }
} same_last_two_bit_with_index;

// function objects for zipwith
class to_double_pair_t : public sketo::functions::base <double_pair(sketo::mindex)> {
public: double_pair operator()(sketo::mindex mi) const { int x = mi.i * size.j + mi.j; return double_pair(x + 1, x - 1); }
} to_double_pair;

class int_pair_with_doule_pair_t : public sketo::functions::base <bool(int_pair, double_pair)> {
public: bool operator()(int_pair x, double_pair y) const { return ((double)x.first == y.first); };
} int_pair_with_double_pair;

// function object for map
class snd_pair_t : public sketo::functions::base <int(int_pair)> {
public: int operator()(int_pair x) const { return x.second; }
} snd_pair;


// get/set of simple data
void test_get_set()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_matrix<int_pair>* as = new sketo::dist_matrix<int_pair>(size);

  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      as->set(sketo::mindex(i, j), to_int_pair(sketo::mindex(i, j)));
    }
  }
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int_pair a = as->get(sketo::mindex(i, j));
      assert(a == to_int_pair(sketo::mindex(i, j)));
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_get_set:OK" << std::endl;
}

// get/set of complex data
void test_get_set2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::dist_matrix<int_vector>* as = new sketo::dist_matrix<int_vector>(size);

  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      as->set(sketo::mindex(i, j), to_int_vector(sketo::mindex(i, j)));
    }
  }
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int_vector a = as->get(sketo::mindex(i, j));
      assert(a == to_int_vector(sketo::mindex(i, j)));
    }
  }
  delete as;
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_get_set2:OK" << std::endl;
}

// gather of simple data
void test_distribute_gather1()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_pair* array = new int_pair[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array[i * size.j + j] = to_int_pair(sketo::mindex(i, j));
    }
  }
  
  sketo::dist_matrix<int_pair> as(size, array);
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      assert(as.get(sketo::mindex(i, j)) == to_int_pair(sketo::mindex(i, j)));
    }
  }
  
  int_pair* array2 = new int_pair[size.j * size.i];
  as.gather(array2);
  if(sketo::rank==0) assert(same_arrays(array, array2, size));
  
  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather1:OK" << std::endl;
}

// distribution from a process, and gather of simple data
void test_distribute_gather2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_pair* array = new int_pair[size.j * size.i];
  sketo::msize diffsize(0, 0);
  if(sketo::rank == 0) {
    for (int i = 0; i < size.i; ++i) {
      for (int j = 0; j < size.j; ++j) {
        array[i * size.j + j] = to_int_pair(sketo::mindex(i, j));
      }
    }
    diffsize = size;
  } else {
    for (int i = 0; i < size.i; ++i) {
      for (int j = 0; j < size.j; ++j) {
        array[i * size.j + j] = to_int_pair(sketo::mindex(0, 0));
      }
    }
  }
  
  sketo::dist_matrix<int_pair> as(diffsize, array, 0);

  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int_pair ip = as.get(sketo::mindex(i, j));
      if(sketo::rank==0) {
        assert(ip == to_int_pair(sketo::mindex(i, j)));
      }
    }
  }
  
  int_pair* array2 = new int_pair[size.j * size.i];
  as.gather(array2);
  if(sketo::rank == 0) assert(same_arrays(array, array2, size));
  
  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather2:OK" << std::endl;
}

// distribution from a process, and gather of complex data
void test_distribute_gather3()
{
  MPI_Barrier(MPI_COMM_WORLD);
  int_vector* array = new int_vector[size.j * size.i];
  if(sketo::rank == 0) {
    for (int i = 0; i < size.i; ++i) {
      for (int j = 0; j < size.j; ++j) {
        array[i * size.j + j] = to_int_vector(sketo::mindex(i, j));
      }
    }
  } else {
    for (int i = 0; i < size.i; ++i) {
      for (int j = 0; j < size.j; ++j) {
        array[i * size.j + j] = to_int_vector(sketo::mindex(0, 0));
      }
    }
  }
  
  sketo::dist_matrix<int_vector> as(size, array, 0);

  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int_vector ip = as.get(sketo::mindex(i, j));
      if(sketo::rank==0) {
        assert(ip == to_int_vector(sketo::mindex(i, j)));
      }
    }
  }
  
  int_vector* array2 = new int_vector[size.j * size.i];
  as.gather(array2);
  if(sketo::rank == 0) assert(same_arrays(array, array2, size));
  
  delete[] array;
  delete[] array2;

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_distribute_gather3:OK" << std::endl;
}

void test_generate()
{
  MPI_Barrier(MPI_COMM_WORLD);
  // skeleton code
  sketo::dist_matrix<int_pair> as;
  as = sketo::matrix_skeletons::generate(size, to_int_pair);
  
  int_pair* array_skel = new int_pair[size.j * size.i];
  as.gather(array_skel);

  // sequential code
  int_pair* array_seq = new int_pair[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = to_int_pair(sketo::mindex(i, j));
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_generate:OK" << std::endl;
}

void test_map()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_matrix<bool> as;
  as = sketo::matrix_skeletons::map(same_last_two_bit,
                                    sketo::matrix_skeletons::generate(size, to_int_pair));
  bool* array_skel = new bool[size.j * size.i];
  as.gather(array_skel);
  
  // sequential code
  bool* array_seq = new bool[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = same_last_two_bit(to_int_pair(sketo::mindex(i, j)));
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map:OK" << std::endl;
}

// void test_map_overwrite()
// {
//   MPI_Barrier(MPI_COMM_WORLD);

//   // skeleton code
//   sketo::dist_matrix<double> as;
//   as = sketo::matrix_skeletons::generate(size, sketo::functions::caster<int, double>());
//   as = sketo::matrix_skeletons::map(std::bind1st(std::multiplies<double>(), 2), as);
//   double* array_skel = new double[size];
//   as.gather(array_skel);

//   // sequential code
//   double *array_seq = new double[size];
//   for (int i = 0; i < size; ++i) {
//     array_seq[i] = 2 * static_cast<double>(i);
//   }

//   if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

//   delete [] array_skel;
//   delete [] array_seq;

//   MPI_Barrier(MPI_COMM_WORLD);
//   sketo::cout << __FILE__ << ":test_map_overwrite:OK" << std::endl;
// }

void test_map_fusion()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_matrix<double> as;
  as = sketo::matrix_skeletons::map(std::bind1st(std::multiplies<double>(), -1),
                                    sketo::matrix_skeletons::map(std::bind1st(std::plus<double>(), -1),
                                                                 sketo::matrix_skeletons::generate(size, doublize)));
}

void test_map_with_index()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_matrix<bool> bs;
  bs = sketo::matrix_skeletons::map_with_index(same_last_two_bit_with_index,
                                               sketo::matrix_skeletons::generate(size, to_int_pair));
  bool* array_skel = new bool[size.i * size.j];
  bs.gather(array_skel);
  
  // sequential code
  bool* array_seq = new bool[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = same_last_two_bit_with_index(sketo::mindex(i, j), to_int_pair(sketo::mindex(i, j)));
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_map_with_index:OK" << std::endl;
}

void test_zip()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  std::pair<int_pair, double_pair> *array_skel = new std::pair<int_pair, double_pair>[size.i * size.j];
  {
    sketo::dist_matrix<std::pair <int_pair, double_pair> > cs;
    cs = sketo::matrix_skeletons::zip(sketo::matrix_skeletons::generate(size, to_int_pair),
                                      sketo::matrix_skeletons::generate(size, to_double_pair));
    cs.gather(array_skel);
  }
      
  // sequential code
  std::pair <int_pair, double_pair> *array_seq = new std::pair <int_pair, double_pair>[size.i * size.j];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j].first = to_int_pair(sketo::mindex(i, j));
      array_seq[i * size.j + j].second = to_double_pair(sketo::mindex(i, j));
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zip:OK" << std::endl;
}

void test_zipwith()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  bool* array_skel = new bool[size.j * size.i];
  {
    sketo::dist_matrix<bool> cs;
    cs = sketo::matrix_skeletons::zipwith(int_pair_with_double_pair,
                                          sketo::matrix_skeletons::generate(size, to_int_pair),
                                          sketo::matrix_skeletons::generate(size, to_double_pair));
    cs.gather(array_skel);
  }

  // sequential code
  bool *array_seq = new bool[size.i * size.j];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = int_pair_with_double_pair(to_int_pair(sketo::mindex(i, j)), to_double_pair(sketo::mindex(i, j)));
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_zipwith:OK" << std::endl;
}

void test_reduce()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int result_skel = sketo::matrix_skeletons::reduce(std::plus<int>(),
                                                    std::plus<int>(),
                                                    sketo::matrix_skeletons::generate(size, intize));

  // sequential code
  int result_seq = 0;
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      result_seq += intize(sketo::mindex(i, j));
    }
  }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce:OK" << std::endl;
}

void test_reduceDC()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int result_skel = sketo::matrix_skeletons::reduce(std::plus<int>(),
                                                    std::plus<int>(),
                                                    sketo::matrix_skeletons::generate(size, intize),
                                                    sketo::matrix_skeletons::ReduceDC());

  // sequential code
  int result_seq = 0;
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      result_seq += intize(sketo::mindex(i, j));
    }
  }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduceDC:OK" << std::endl;
}

void test_scan()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  int* array_skel = new int[size.i * size.j];
  {
    sketo::dist_matrix<int> as = sketo::matrix_skeletons::scan(std::plus<int>(),
                                                               std::plus<int>(),
                                                               sketo::matrix_skeletons::generate(size, intize));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size.i * size.j];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = intize(sketo::mindex(i, j));
    }
  }
  for (int i = 0; i < size.i; ++i) {
    for (int j = 1; j < size.j; ++j) {
      array_seq[i * size.j + j] = array_seq[i * size.j + j-1] + array_seq[i * size.j + j];
    }
  }
  for (int j = 0; j < size.j; ++j) {
    for (int i = 1; i < size.i; ++i) {
      array_seq[i * size.j + j] = array_seq[(i-1) * size.j + j] + array_seq[i * size.j + j];
    }
  }
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan:OK" << std::endl;
}

void test_scanr()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  int* array_skel = new int[size.i * size.j];
  {
    sketo::dist_matrix<int> as = sketo::matrix_skeletons::scanr(std::plus<int>(),
                                                                std::plus<int>(),
                                                                sketo::matrix_skeletons::generate(size, intize));
    as.gather(array_skel);
  }

  // sequential code
  int *array_seq = new int[size.i * size.j];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = intize(sketo::mindex(i, j));
    }
  }
  for (int i = 0; i < size.i; ++i) {
    for (int j = size.j - 2; j >= 0; j--) {
      array_seq[i * size.j + j] = array_seq[i * size.j + j] + array_seq[i * size.j + j+1];
    }
  }
  for (int j = 0; j < size.j; ++j) {
    for (int i = size.i - 2; i >= 0; i--) {
      array_seq[i * size.j + j] = array_seq[i * size.j + j] + array_seq[(i+1) * size.j + j];
    }
  }
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr:OK" << std::endl;
}

void test_reduce2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_mat result_skel = sketo::matrix_skeletons::reduce(im_abv,
                                                        im_bsd,
                                                        sketo::matrix_skeletons::generate(size, im_single));

  // sequential code
  int_mat result_seq;
  for (int i = 0; i < size.i; ++i) {
    int_mat iv;
    for (int j = 0; j < size.j; ++j) {
      if(iv.size()==0) iv = im_single(sketo::mindex(i, j));
      else iv = im_bsd(iv, im_single(sketo::mindex(i, j)));
    }
    if(result_seq.size()==0) result_seq = iv;
    else result_seq = im_abv(result_seq, iv);
  }
  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduce2:OK" << std::endl;
}

void test_reduceDC2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  int_mat result_skel = sketo::matrix_skeletons::reduce(im_abv,
                                                        im_bsd,
                                                        sketo::matrix_skeletons::generate(size, im_single),
                                                        sketo::matrix_skeletons::ReduceDC());

  // sequential code
  int_mat result_seq;
  for (int i = 0; i < size.i; ++i) {
    int_mat iv;
    for (int j = 0; j < size.j; ++j) {
      if(iv.size()==0) iv = im_single(sketo::mindex(i, j));
      else iv = im_bsd(iv, im_single(sketo::mindex(i, j)));
    }
    if(result_seq.size()==0) result_seq = iv;
    else result_seq = im_abv(result_seq, iv);
  }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_reduceDC2:OK" << std::endl;
}

void test_scan2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  int_mat* array_skel = new int_mat[size.i * size.j];
  {
    sketo::dist_matrix<int_mat> as = sketo::matrix_skeletons::scan(im_abv,
                                                                   im_bsd,
                                                                   sketo::matrix_skeletons::generate(size, im_single));
    as.gather(array_skel);
  }

  // sequential code
  int_mat *array_seq = new int_mat[size.i * size.j];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = im_single(sketo::mindex(i, j));
    }
  }
  for (int i = 0; i < size.i; ++i) {
    for (int j = 1; j < size.j; ++j) {
      array_seq[i * size.j + j] = im_bsd(array_seq[i * size.j + j-1], array_seq[i * size.j + j]);
    }
  }
  for (int j = 0; j < size.j; ++j) {
    for (int i = 1; i < size.i; ++i) {
      array_seq[i * size.j + j] = im_abv(array_seq[(i-1) * size.j + j], array_seq[i * size.j + j]);
    }
  }
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scan2:OK" << std::endl;
}

void test_scanr2()
{
  MPI_Barrier(MPI_COMM_WORLD);
  
  // skeleton code
  int_mat* array_skel = new int_mat[size.i * size.j];
  {
    sketo::dist_matrix<int_mat> as = sketo::matrix_skeletons::scanr(im_abv,
                                                                im_bsd,
                                                                sketo::matrix_skeletons::generate(size, im_single));
    as.gather(array_skel);
  }

  // sequential code
  int_mat *array_seq = new int_mat[size.i * size.j];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = im_single(sketo::mindex(i, j));
    }
  }
  for (int i = 0; i < size.i; ++i) {
    for (int j = size.j - 2; j >= 0; j--) {
      array_seq[i * size.j + j] = im_bsd(array_seq[i * size.j + j], array_seq[i * size.j + j+1]);
    }
  }
  for (int j = 0; j < size.j; ++j) {
    for (int i = size.i - 2; i >= 0; i--) {
      array_seq[i * size.j + j] = im_abv(array_seq[i * size.j + j], array_seq[(i+1) * size.j + j]);
    }
  }
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_scanr2:OK" << std::endl;
}



// shift ow
template<int di, int dj>
void test_shift1()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_matrix<int_pair> as;
  as = sketo::matrix_skeletons::shift<di,dj>(shift_fun_identity,
                                             sketo::matrix_skeletons::generate(size, to_int_pair));
  int_pair* array_skel = new int_pair[size.j * size.i];
  as.gather(array_skel);

  // sequential code
  int_pair* array_seq_0 = new int_pair[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq_0[i * size.j + j] = to_int_pair(sketo::mindex(i, j));
    }
  }
  int_pair* array_seq = new int_pair[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int idi = i - di;
      int jdj = j - dj;
      int ii = ((idi % size.i) + size.i) % size.i; 
      int jj = ((jdj % size.j) + size.j) % size.j;
      array_seq[i * size.j + j] = array_seq_0[ii * size.j + jj];
      if(idi!=ii || jdj!=jj) {
        array_seq[i * size.j + j] = shift_fun_identity(sketo::mindex(idi, jdj), array_seq[i * size.j + j]);   
      }
    }
  }

  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  delete [] array_seq_0;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shift1<" << di << ", " << dj << ">:OK" << std::endl;
}


// shift fusion
template<int di, int dj>
void test_shift2()
{
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  sketo::dist_matrix<int> as;
  as = sketo::matrix_skeletons::map(snd_pair, 
                                    sketo::matrix_skeletons::shift<di,dj>(shift_fun_identity,
                                                                          sketo::matrix_skeletons::generate(size, to_int_pair)));
  int* array_skel = new int[size.j * size.i];
  as.gather(array_skel);

// sequential code
  int_pair* array_seq_0 = new int_pair[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq_0[i * size.j + j] = to_int_pair(sketo::mindex(i, j));
    }
  }
  int_pair* array_seq_1 = new int_pair[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int idi = i - di;
      int jdj = j - dj;
      int ii = ((idi % size.i) + size.i) % size.i; 
      int jj = ((jdj % size.j) + size.j) % size.j;
      array_seq_1[i * size.j + j] = array_seq_0[ii * size.j + jj];
      if(idi!=ii || jdj!=jj) {
        array_seq_1[i * size.j + j] = shift_fun_identity(sketo::mindex(idi, jdj), array_seq_1[i * size.j + j]);   
      }
    }
  }
  int * array_seq = new int[size.j * size.i];
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      array_seq[i * size.j + j] = snd_pair(array_seq_1[i * size.j + j]);
    }
  }
  // compare
  if (sketo::is_master()) assert(same_arrays(array_skel, array_seq, size));

  delete [] array_skel;
  delete [] array_seq;
  delete [] array_seq_0;
  delete [] array_seq_1;
  
  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_shift2<" << di << ", " << dj << ">:OK" << std::endl;
}

#ifdef __GXX_EXPERIMENTAL_CXX0X__  // using c++0x
void test_lambda() {
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  auto plus = [](int a, int b) { return a + b; };
  int result_skel = sketo::matrix_skeletons::reduce(
    plus, plus,
    sketo::matrix_skeletons::map(
      [](int a) { return a * a; },
      sketo::matrix_skeletons::generate(
        size,
        [](sketo::mindex mi) { return 1 + 0*(mi.i * size.j + mi.j); })));
  
  // sequential code
  int result_seq = 0;
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int k = intize(sketo::mindex(i, j));
      result_seq += k * k;
    }
  }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_lambda:OK" << std::endl;
}
void test_auto_fusion() {
  MPI_Barrier(MPI_COMM_WORLD);

  // skeleton code
  auto plus = [](int a, int b) { return a + b; };
  int result_skel = sketo::matrix_skeletons::reduce(
    plus, plus,
    sketo::matrix_skeletons::map(
      [](int a) { return a * a; },
      sketo::matrix_skeletons::generate(
        size,
        [](sketo::mindex mi) { return 1 + 0*(mi.i * size.j + mi.j); })));
  
  // sequential code
  int result_seq = 0;
  for (int i = 0; i < size.i; ++i) {
    for (int j = 0; j < size.j; ++j) {
      int k = intize(sketo::mindex(i, j));
      result_seq += k * k;
    }
  }

  // compare
  if (sketo::is_master()) assert(result_skel == result_seq);

  MPI_Barrier(MPI_COMM_WORLD);
  sketo::cout << __FILE__ << ":test_auto_fusion:OK" << std::endl;
}

#endif




int sketo::main(int, char**)
{
  test_get_set();
  test_get_set2();
  test_distribute_gather1();
  test_distribute_gather2();
  test_distribute_gather3();

  test_generate();
  test_map();
  test_map_fusion();
  test_map_with_index();
  test_zip();
  test_zipwith();
  test_reduce();
  test_reduceDC();
  test_scan();
  test_scanr();
  test_reduce2();
  test_reduceDC2();
  test_scan2();
  test_scanr2();
  test_shift1<1, 2>();
  test_shift1<3, 0>();
  test_shift1<0, 4>();
  test_shift1<-1, -2>();
  test_shift1<-3, 0>();
  test_shift1<0, -4>();
  test_shift1<-30, 0>();
  test_shift1<0, -40>();
  test_shift1<-30, -40>();
  // test_shift1<1, 0>();
  // test_shift1<2, 0>();
  // test_shift1<3, 0>();
  // test_shift1<4, 0>();
  // test_shift1<5, 0>();
  // test_shift1<6, 0>();
  // test_shift1<7, 0>();
  // test_shift1<8, 0>();
  // test_shift1<9, 0>();
  // test_shift1<10, 0>();
  // test_shift1<11, 0>();
  // test_shift1<12, 0>();
  // test_shift1<13, 0>();
  // test_shift1<14, 0>();
  test_shift2<1, 2>();
  test_shift2<1, 2>();
  test_shift2<3, 0>();
  test_shift2<0, 4>();
  test_shift2<-1, -2>();
  test_shift2<-3, 0>();
  test_shift2<0, -4>();
  test_shift2<-30, 0>();
  test_shift2<0, -40>();
  test_shift2<-30, -40>();
 
// to test the lambda expression, we should compile the test program by hand:
// /usr/local/mpich2-1.2.1p1/bin//mpicxx -L../src -lsketo -I../include -O3 test_matrix_skeletons.cpp -o test_matrix_skeletons.sketo -std=gnu++0x

#ifdef __GXX_EXPERIMENTAL_CXX0X__  // using c++0x
  test_lambda();
#endif
  return 0;
}
